/* 
 * ++/Algo/BBAlgo.h
 * 
 * This source include the data type for the Branch and Bound Algo level.
 * This the unit of task scheduled by the Kernel class.
 */

#ifndef BOBBBALGO_ABSTRACT
#define BOBBBALGO_ABSTRACT

namespace  Bob {

/** @defgroup BBAlgoGRP The Branch and Bound Algorithm.
  *  @brief This module describe all classes to implement a Branch and Bound  algorithm.
  *  @ingroup SchedAlgoGRP
  *  @{

Each classes use to implements Branch and Bound inherits from their 
associated classes of the module @ref SchedAlgoGRP.


For example, we could solve My Problem  using this method,
We give step by step the specific classes and their associated methods and attributs.

  -# Define the MyTrait class with the forward definition of the classes.
     \verbatim
#include<bobpp/bobpp>

class MyProblem;
class MyInstance;
class MyGenChild;
class MyNode;


class MyTrait {
  public:
    typedef MyNode Node;
    typedef MyInstance Instance;
    typedef MyGenChild GenChild;
    typedef Bob::BBAlgo<MyTrait> Algo;
    typedef Bob::BBGoalBest<MyTrait> Goal;
    typedef Bob::BestEPri<MyNode> PriComp;
    typedef Bob::BBStat Stat;
};
     \endverbatim


  -# define the MyNode class
     \verbatim
     class MyNode : public Bob::BBIntMinNode {
        // specific datas
        public:
                MyNode() : Bob::BBIntMinNode(0) {
                }
                MyNode(int n) : Bob::BBIntMinNode(n) {
                }
                virtual ~MyNode() {
                }

};
     \endverbatim
     To fixe the evaluation of a node you could use the methods for the Eval in the 
     Eval class (setEval(T t) and T getEval()).
     
     This MyNode inherits form the BBIntMaxNode, but could also inherits
     from the BBLongMinNode, etc... @ref BBAlgoSHGRP
     You could also inherit from the BBNode<T,Compare> where T is a type for a cost
     and Compare is a class to Compare these cost and could be Min or Max.

  -# defines the MyGenChild
     \verbatim
class MyGenChild: public Bob::BBGenChild<MyTrait> {
        public:
                /// Constructor of MyGenChild
                MyGenChild(MyTrait::Instance *_inst,MyTrait::Algo *_al) :
                        Bob::BBGenChild<MyTrait>(_inst,_al) {
                        }
                virtual ~MyGenChild() {
                }
                /// Operator called by the algo to Generate child
                void operator()(MyNode *n) {
                        int nbs=inst->getnbson();
                        int step=inst->getstep();
                        int i;
                        for (i=0;i<nbs;i++ ) {
                                MyNode *ns=new MyNode(n->getEval()+(rand()%(step-1))+1);
                                algo->StEval();
                                algo->Search(ns);
                        }
                }
};

     \endverbatim
     When you call a Search(n) with n->IsSol() that returns true, n is not
     reinserted in the GPQ.
     This semantic could be a problem, for some application. For example when solving
     a Knapsac problem using a Costed divide and Conquer algorithms, one algorithm
     is at the level i to create two child nodes, one where the object i has been added 
     in the bag and one where the object i is not added in the bag.
     Thus a node could is a solution and a subproblem at the same time.
     you could see the sources of the knapsac application to see an example to solve
     this problem.

  -# defines the MyInstance
     \verbatim
class MyInstance : public Bob::BBInstance<MyTrait> {
   // specific datas for the instance
        public:
                /// Constructor
                MyInstance() : Bob::BBInstance<MyTrait>() {
                }
                void InitAlgo(MyTrait::Algo *a) const {
                        std::cout<<"MyInstance::Init(Algo)"<<std::endl;
                        MyNode *nd= new MyNode(lb);
         /// Initialize de data for the root node
                        a->GPQIns(nd);
                }
                /// Get the initial best known solution
                MyNode *getSol() { 
         return new MyNode(ub);
      }
};

         \endverbatim

     You could have a look to the Example soource code that comes with the Bob++ Library.
*/

/** The Branch and Bound node
    a Node of a Branch is evaluated.
 */
template <class T, class Compare>
class BBNode : public Eval<T,Compare> , public SchedNode {
public :
   /// Constructor
   BBNode() : Eval<T,Compare>() ,SchedNode(){ }
   /// Constructor
   BBNode(T v) : Eval<T,Compare>(v) ,SchedNode(){ }
   /// Constructor
   BBNode(BBNode<T,Compare> *bn) : Eval<T,Compare>(bn->getEval()) ,SchedNode(*bn){ }
   /// Constructor
   BBNode(BBNode<T,Compare> &bn) : Eval<T,Compare>(bn.getEval()) ,SchedNode(bn){ }
   /// Constructor
   BBNode(const BBNode<T,Compare> &bn) : Eval<T,Compare>(bn.getEval()) ,SchedNode(bn){ }
   /// Destructor
   virtual ~BBNode() {}
   /** copy method
     */
   virtual void copy(const BBNode<T,Compare> &tn) {
         SchedNode::copy(tn);
         Eval<T,Compare>::copy(tn);
   }
   /// print the node information
   virtual std::ostream &Prt(std::ostream &os) const {
        SchedNode::Prt(os);
        os << "Eval:"<<Eval<T,Compare>::getEval()<<std::endl;
        return os;
   }
   /** Fill the strbuff with node information
     * @param st the strbuff to fill.
     */
   virtual void log_space(strbuff<> &st) {
      SchedNode::log_space(st);
      st << ","<<Eval<T,Compare>::getEval();
   } 
   /** Fill the strbuff with node information
     * @param st the strbuff to fill.
     */
   virtual void lab_log_space(strbuff<> &st) {
      SchedNode::log_space(st);
      st << ","<<"evaluation";
   } 

   /// Pack method to serialize the Node
   virtual void Pack(Serialize &bs)  const {
        SchedNode::Pack(bs);
        Eval<T,Compare>::Pack(bs);
   }
   /// Unpack method to deserialize the Node
   virtual void UnPack(DeSerialize &bs)  {
        SchedNode::UnPack(bs);
        Eval<T,Compare>::UnPack(bs);
   }

};

/** The Branch and Bound Instance Algorithm
 */
template <class Trait>
class BBInstance : public SchedInstance<Trait>  {
public :
   /// type of a node.
   typedef typename Trait::Node TheNode;
   /// type of the algorithm.
   typedef typename Trait::Algo TheAlgo;
   /// type of the algorithm.
   typedef typename Trait::Instance TheInstance;
   /// type of evaluation for the node.
   typedef typename TheNode::EvalType EvalType;
   /** Constructor
     * @param _n an std::string that stores the name of the instance
     */
   BBInstance(std::string &_n) : SchedInstance<Trait>(_n) {
   }
   /** Constructor
     * @param _n a char * that stores the name of the instance
     */
   BBInstance(const char *_n) : SchedInstance<Trait>(_n) {
   }
   /** Constructor
     * with no parameter the name is "??"
     */
   BBInstance() : SchedInstance<Trait>(std::string("??")) {
   }
   /** Constructor with copy of a BBInstance
     * @param bbi the BBInstance to copy
     */
   BBInstance(const BBInstance &bbi) : SchedInstance<Trait>(bbi) {
   }
   /// Destructor
   virtual ~BBInstance() {}
   /// Get the solution
   virtual TheNode *getSol()=0;

   /// Pack method to serialize the Node
   virtual void Pack(Serialize &bs) const {
        SchedInstance<Trait>::Pack(bs);
   }
   /// Unpack method to deserialize the Node
   virtual void UnPack(DeSerialize &bs)  {
        SchedInstance<Trait>::UnPack(bs);
   }
};


/**
 * The Statistics for Branch and Bound algorithm
 */
class BBStat : public SchedStat {
   public:
      /// Constructor
      BBStat(Id *s,PQStat *_pqs,GoalStat *_gst) :
           SchedStat('B',s,_pqs,_gst) {
         add_counter('e',"Evaluated Nodes");
      }
      /// Destructor
      virtual ~BBStat() {}
      /// Counts the number of evaluated nodes
      void Eval() { get_counter('e')++; }
};

/** The Branch and Bound GenChild
 */
template <class Trait>
class BBGenChild : public SchedGenChild<Trait> {
   public:
      /// Type of a Node
      typedef typename Trait::Node TheNode;
      /// Type of the Info
      typedef typename Trait::Instance TheInstance;
      /// Type for the goal of the search
      typedef typename Trait::Goal TheGoal;
      /// Type of the Algo
      typedef typename Trait::Algo TheAlgo;
      /// Constructor
      BBGenChild(const TheInstance *_inst, TheAlgo*_al) : SchedGenChild<Trait>(_inst,_al) { }
      /// Destructor
      virtual ~BBGenChild() {}
};

/** @defgroup BBGoal The Different Branch and Bound Goals
 * A Bob++ Algo coud be instancied with one goal or another.
 * The goal must be defined in the Trait class
 * @{
 */

/** The Classical Branch and Bound Goal.
 * This goal is the classical Branch and Bound goal where the aim of the 
 * search is to get the best solution.
 */
template<class Trait>
class BBGoalBest :public SchedGoal<Trait> {
  public :
    /// Type of a Node
    typedef typename Trait::Node TheNode;
    /// Type of a Node
    typedef typename TheNode::EvalType  TheType;
    /// Type of the Instance
    typedef typename Trait::Instance TheInstance;
    /// Type for the goal of the search
    typedef typename Trait::Goal TheGoal;
    /// Type of the Statistics 
    typedef typename Trait::Stat TheStat;
  protected:
    /// the solution
    TheNode *sol;
  public:
   /// Constructor
   BBGoalBest(bool g=false): SchedGoal<Trait>(g), sol(0) { }
   /// Constructor
   BBGoalBest(BBGoalBest<Trait> &bb): SchedGoal<Trait>(bb), sol(0) { 
      std::cout << "BBGoalBest Copy &bb\n";
      if ( bb.sol!=0 ) { sol=bb.sol; sol->addRef(); }
   }
   /// Constructor
   BBGoalBest(const BBGoalBest<Trait> &bb): SchedGoal<Trait>(bb), sol(0) { 
      std::cout << "BBGoalBest Copy const &bb\n";
      if ( bb.sol!=0 ) { sol=bb.sol; sol->addRef(); }
   }
   /// Destructor
   virtual ~BBGoalBest() {
         if ( sol!=0 ) {
            sol->remRef();
            if ( sol->isDel() ) { delete sol; sol=0;}
         }
   }
   /// get the value of the best solution
   TheType getBest() { 
      if ( sol ==0 ) return 0;
      return sol->getEval();
   }
   /** Intialize the Goal with the given instance.
     * @param ti the instance.
     */
   virtual void Init(TheInstance *ti) {
      sol=ti->getSol();
      //std::cout << "Node Sol:"<<*sol<<std::endl;
      sol->addRef();
   }
    /** Method called when a node is a solution
      * @param n the solution node
      */ 
    virtual bool updSol(TheNode *n) {
       if ( n==0 ) return false;
       DBGAFF_ALGO("BBGoalBest::updsol()","Node is a solution");
       SchedGoal<Trait>::getStat()->sol();
       if ( sol==0 || (*n)>(*sol) ) {
          DBGAFF_ALGO("BBGoalBest::updsol()","A Better Solution is found");
          if ( sol !=0 ) {
             sol->remRef();
             if ( sol->isDel() ) delete sol;
          }
          SchedGoal<Trait>::getStat()->upd();
          sol=n;
          sol->addRef();
          if (core::dispBestVal())
            std::cout << "Current best solution: " << sol->getEval() <<
                std::endl;
          return true;
       }
       return false;
    }

   /** Method to test if the node is suitable to be explore later. 
     * @param n the node to test
     */
   virtual bool is4Search(TheNode *n) {
      DBGAFF_ALGO("BBGoalBest::is4Search()","Node is a subproblem");
      if ( sol==0  ) {
          // std::cout<<"isSearch Test NULL with "<<n->getEval()<<"\n";
         return true;
      }
      if ( (*n)>(*sol) ) {
          // std::cout<<"isSearch Test "<<sol->getEval()<<" with "<<n->getEval()<<"\n";
         return true;
      } 
      return false;
   }
   /** the merging method is used to update the goal with another one.
     * this method is used in distributed environment.
     * @param _tg TheGoal used to updated the Goal.
     */
   virtual bool current_merge(const TheGoal *_tg) {
      if ( _tg->sol==0 ) return false;
      return updSol(_tg->sol);
   }

   /** the merging method is used to update the goal with another one.
     * this method is used in distributed environment.
     * @param _tg TheGoal used to updated the Goal.
     */
   virtual bool final_merge(const TheGoal *_tg) {
      if ( _tg->sol==0 ) return false;
      return updSol(_tg->sol);
   }

   /** Return the best known solution */
   virtual const TheNode * best() {
     return sol;
   }

   /** Return the worst known solution */
   virtual const TheNode * worst() {
     return sol;
   }

   /// Prt Method display the Solution data structure
   virtual ostream &Prt(ostream &os) const {
       os<<"Solution-------------";
       if ( sol==0 ) {
          os <<"NULL\n";
          return os;
       }
       return sol->Prt(os);
   }
   /** Pack method to Serialize the object
     * @param bs the DeSerialize object
     */
   virtual void Pack(Serialize &bs) const {
       int i=0;
       DBGAFF_ENV("BBGoalBest::Pack","Packing Node");
       SchedGoal<Trait>::Pack(bs);
       if ( sol==0 ) {
         //std::cout <<" Pack no sol\n";
         bs.Pack(&i,1);
         return;
       }
       i=1;
       bs.Pack(&i,1);
       sol->Pack(bs);
       //sol->Prt(std::cout);
   }
   /** Unpack method to deserialize the 
     * @param bs the DeSerialize object
     */
   virtual void UnPack(DeSerialize &bs)  {
       int i;
       DBGAFF_ENV("BBGoalBest::UnPack","Unpacking Node");
       SchedGoal<Trait>::UnPack(bs);
       bs.UnPack(&i,1);
       if ( i==0 ) { 
         sol=0; 
         //std::cout <<" UnPack no sol\n";
         return;
       }
       sol = new TheNode();
       sol->UnPack(bs);
       //sol->Prt(std::cout);
   }

};


/** The Goal that counts the number of best solution.
 * This goal is quite different. The aim of the search is to count
 * the number of solution with the best cost.
 */
template<class Trait>
class BBGoalCount : public SchedGoal<Trait> {
  public:
    /// Type of a Node
    typedef typename Trait::Node TheNode;
    /// Type of the evaluation of the node
    typedef typename TheNode::EvalType  TheType;
    /// Type of the Info
    typedef typename Trait::Instance TheInstance;
     /// Type for the goal of the search
    typedef typename Trait::Goal TheGoal;
  protected:
    /// the solution
    TheNode *sol;
    /// the number of solution
    int nbsol;
  public:
    /// Constructor
    BBGoalCount(bool g=false) : SchedGoal<Trait>(g),sol(0),nbsol(0) {   }
    /// Constructor
    BBGoalCount(BBGoalCount<Trait> &bb) :SchedGoal<Trait>(bb),sol(0),nbsol(bb.nbsol){
      if ( bb.sol!=0 ) { sol=bb.sol; sol->addRef(); }
    }
    /// Constructor
    BBGoalCount(const BBGoalCount<Trait> &bb) : SchedGoal<Trait>(bb),sol(0),
      nbsol(bb.nbsol) {   
      if ( bb.sol!=0 ) { sol=bb.sol; sol->addRef(); }
    }
    /// Destructor
    virtual ~BBGoalCount() { 
         if ( sol!=0 ) {
            sol->remRef();
            if ( sol->isDel() ) { delete sol; sol=0;}
         }
    }
    /** Intialize the Goal.
      * @param ti the Instance used to initialize the Goal
      *
      */
    virtual void Init(TheInstance *ti) {
         sol=ti->getSol();
         nbsol=sol->getSol(); 
         sol->addRef();
    }
    /// get the value of the best solution
    TheType getBest() { 
      if ( sol ==0 ) return 0;
      return sol->getEval();
    }

    /** Method called when a node is a solution
      * @param n the solution node
      */ 
    virtual bool updSol(TheNode *n) {
       if ( n==0 ) return false;
       DBGAFF_ALGO("BBGoalCount::updSol()","Node is a solution");
 	    if ( sol==0 || (*n)>(*sol) ) {
          DBGAFF_ALGO("BBGoalCount::updSol()","A Better Solution is found");
          if ( sol !=0 ) {
             sol->remRef();
             if ( sol->isDel() ) delete sol;
          }
          sol=n;
          nbsol=sol->getSol();
          sol->addRef();
          SchedGoal<Trait>::getStat()->sol();
          if (core::dispBestVal())
            std::cout << "Current best solution: " << sol->getEval() <<
                std::endl;
          return true;
       }
 	    if ( (*n)==(*sol) ) {
          DBGAFF_ALGO("BBGoalCount::updSol()","A as good Solution is found");
          SchedGoal<Trait>::getStat()->sol();
          nbsol+=sol->getSol();
       }
       return false;
    }
    /** Method to test if the node is suitable to be explore later. 
      * @param n the node to test
      */
    virtual bool is4Search (TheNode * n) {
      //std::cout<<"isSearch Test "<<(sol==0?"NULL":sol->getEval())<<" with "<<n->getEval()<<"\n";
      DBGAFF_ALGO("BBGoalCount::is4Search()","Node is a subproblem");
      if ( sol==0 || (*n)>(*sol) || (*n)==(*sol) ) 
         return true;
      return false;
    }
    /** The final merging method is used to update the goal with another one.
      * but at the end of the execution of the algorithm
      * this method is used in distributed environment.
      * @param _tg TheGoal used to updated the Goal.
      */
    virtual bool final_merge(const TheGoal *_tg) {
      if ( _tg->sol==0 ) return false;
      return updSol(_tg->sol);
    }

    /** The current merging method is used to update the goal with another 
      * one, during the execution of the associated algorithm.
      * this method is used in distributed environment.
      * @param _tg TheGoal used to updated the Goal.
      */
    virtual bool current_merge(const TheGoal *_tg) {
      if ( _tg->sol==0 ) return false;
      return updSol(_tg->sol);
    }

   /** Return the best known solution */
   virtual const TheNode * best() {
     return sol;
   }

   /** Return the worst known solution */
   virtual const TheNode * worst() {
     return sol;
   }

   /** Pack method to Serialize the object
     * @param bs the Serialize object
     */
   virtual void Pack(Serialize &bs) const {
       int i=0;
       DBGAFF_ENV("BBGoalCount::Pack","Packing Node");
       SchedGoal<Trait>::Pack(bs);
       if ( sol==0 ) {
         bs.Pack(&i,1);
         return;
       }
       i=1;
       bs.Pack(&i,1);
       bs.Pack(&nbsol,1);
       sol->Pack(bs);
   }
   /** Unpack method to deserialize the SchedNode
     * @param bs the DeSerialize object
     */
   virtual void UnPack(DeSerialize &bs)  {
       int i;
       DBGAFF_ENV("BBGoalCount::UnPack","Unpacking Node");
       SchedGoal<Trait>::UnPack(bs);
       bs.UnPack(&i,1);
       if ( i==0 ) sol=0; return;
       bs.UnPack(&nbsol,1);
       sol = new TheNode();
       sol->UnPack(bs);
   }
    /// Prt Method display the Solution data structure
    virtual ostream &Prt(ostream &os) const {
       if ( sol==0 ) 
          os<<"Solution------------- NULL #:"<<nbsol<<"\n";
       else
          os<<"Solution-------------"<< sol->getEval() <<"#:"<<nbsol<<"\n";
       return os;
    }
};

/** The Goal to test a search best solution.
 * This goal yields to an exhaustive search
 */
template<class Trait>
class BBGoalAll : public SchedGoal<Trait> {
  public:
    /// Type of a Node
    typedef typename Trait::Node TheNode;
    /// Type of the evaluation of the node
    typedef typename TheNode::EvalType  TheType;
    /// Type of the Info
    typedef typename Trait::Instance TheInstance;
     /// Type for the goal of the search
    typedef typename Trait::Goal TheGoal;
  protected:
    /// the solution
    TheNode *sol;
    /// the number of solution
    int nbsol;
  public:
    /// Constructor
    BBGoalAll(bool g=false) : SchedGoal<Trait>(g),sol(0),nbsol(0) {   }
    /// Constructor
    BBGoalAll(BBGoalAll<Trait> &bb) :SchedGoal<Trait>(bb),sol(0),nbsol(bb.nbsol){
      if ( bb.sol!=0 ) { sol=bb.sol; sol->addRef(); }
    }
    /// Constructor
    BBGoalAll(const BBGoalAll<Trait> &bb) : SchedGoal<Trait>(bb),sol(0),
      nbsol(bb.nbsol) {   
      if ( bb.sol!=0 ) { sol=bb.sol; sol->addRef(); }
    }
    /// Destructor
    virtual ~BBGoalAll() { 
         if ( sol!=0 ) {
            sol->remRef();
            if ( sol->isDel() ) { delete sol; sol=0;}
         }
    }
    /** Intialize the Goal.
      * @param ti the Instance used to initialize the Goal
      *
      */
    virtual void Init(TheInstance *ti) {
         sol=ti->getSol();
         nbsol=sol->getSol(); 
         sol->addRef();
    }
    /// get the value of the best solution
    TheType getBest() { 
      if ( sol ==0 ) return 0;
      return sol->getEval();
    }

    /** Method called when a node is a solution
      * @param n the solution node
      */ 
    virtual bool updSol(TheNode *n) {
       if ( n==0 ) return false;
       DBGAFF_ALGO("BBGoalCount::updSol()","Node is a solution");
 	    if ( sol==0 || (*n)>(*sol) ) {
          DBGAFF_ALGO("BBGoalCount::updSol()","A Better Solution is found");
          if ( sol !=0 ) {
             sol->remRef();
             if ( sol->isDel() ) delete sol;
          }
          sol=n;
          nbsol=sol->getSol();
          sol->addRef();
          SchedGoal<Trait>::getStat()->sol();
          if (core::dispBestVal())
            std::cout << "Current best solution: " << sol->getEval() <<
                std::endl;
          return true;
       }
 	    if ( (*n)==(*sol) ) {
          DBGAFF_ALGO("BBGoalCount::updSol()","A as good Solution is found");
          SchedGoal<Trait>::getStat()->sol();
          nbsol+=sol->getSol();
       }
       return false;
    }
    /** Method to test if the node is suitable to be explore later. 
      * @param n the node to test
      */
    virtual bool is4Search (TheNode * n) {
      //std::cout<<"isSearch Test "<<(sol==0?"NULL":sol->getEval())<<" with "<<n->getEval()<<"\n";
      DBGAFF_ALGO("BBGoalCount::is4Search()","Node is a subproblem");
      if ( sol==0 || (*n)>(*sol) || (*n)==(*sol) ) 
         return true;
      return true;
    }
    /** The final merging method is used to update the goal with another one.
      * but at the end of the execution of the algorithm
      * this method is used in distributed environment.
      * @param _tg TheGoal used to updated the Goal.
      */
    virtual bool final_merge(const TheGoal *_tg) {
      if ( _tg->sol==0 ) return false;
      return updSol(_tg->sol);
    }

    /** The current merging method is used to update the goal with another 
      * one, during the execution of the associated algorithm.
      * this method is used in distributed environment.
      * @param _tg TheGoal used to updated the Goal.
      */
    virtual bool current_merge(const TheGoal *_tg) {
      if ( _tg->sol==0 ) return false;
      return updSol(_tg->sol);
    }

   /** Return the best known solution */
   virtual const TheNode * best() {
     return sol;
   }

   /** Return the worst known solution */
   virtual const TheNode * worst() {
     return sol;
   }

   /** Pack method to Serialize the object
     * @param bs the Serialize object
     */
   virtual void Pack(Serialize &bs) const {
       int i=0;
       DBGAFF_ENV("BBGoalCount::Pack","Packing Node");
       SchedGoal<Trait>::Pack(bs);
       if ( sol==0 ) {
         bs.Pack(&i,1);
         return;
       }
       i=1;
       bs.Pack(&i,1);
       bs.Pack(&nbsol,1);
       sol->Pack(bs);
   }
   /** Unpack method to deserialize the SchedNode
     * @param bs the DeSerialize object
     */
   virtual void UnPack(DeSerialize &bs)  {
       int i;
       DBGAFF_ENV("BBGoalCount::UnPack","Unpacking Node");
       SchedGoal<Trait>::UnPack(bs);
       bs.UnPack(&i,1);
       if ( i==0 ) sol=0; return;
       bs.UnPack(&nbsol,1);
       sol = new TheNode();
       sol->UnPack(bs);
   }
    /// Prt Method display the Solution data structure
    virtual ostream &Prt(ostream &os) const {
       if ( sol==0 ) 
          os<<"Solution------------- NULL #:"<<nbsol<<"\n";
       else
          os<<"Solution-------------"<< sol->getEval() <<"#:"<<nbsol<<"\n";
       return os;
    }
};

/**
 * @}
 */


/** 
 * The Branch and Bound Algorithm
 */
template <class Trait>
class BBAlgo : public SchedAlgo<Trait> {
   public:
      /// Type of the node
      typedef typename Trait::Node TheNode;
      /// Type of the info 
      typedef typename Trait::Instance TheInstance;
      /// Type of the priotiy comparator
      typedef typename Trait::PriComp ThePriComp;
      /// Type of the Statistics 
      typedef typename Trait::Stat TheStat;
      /// Type of the priotiy comparator 
      typedef typename Trait::GenChild TheGenChild;
      /// Type for the goal of the search
      typedef typename Trait::Goal TheGoal;
      /// Type of the Algo
      typedef typename Trait::Algo TheAlgo;
      /// Type of the pq
      typedef PQInterface<TheNode,ThePriComp> ThePQ;
      /** Constructor
       * @param i the instance to solve
       * @param pq the Priority queue
       * @param tg the goal of the search
       * @param st the statistics object
       */
      BBAlgo(const TheInstance *i,ThePQ *pq, TheGoal *tg,TheStat *st) : 
	                SchedAlgo<Trait>(i,pq,tg,st) {
      }
      /// destructor
      virtual ~BBAlgo(){ 
      }
      /** 
       @name additional Branch and Bound Statistics
      */
      //@{
      /// statisitics for evaluated nodes
      void StEval() { (SchedAlgo<Trait>::getStat())->Eval(); }
      //@}

};

/** @defgroup BBAlgoSHGRP The shortcuts definition for Branch and Bound Algorithm
    @brief This modules presents shortcuts definition of the Instance/Algo/Node to 
   implement a Branch and Bound 
 */
//@{
/// @name Minimization and integer evaluation
//@{
typedef BBNode<int,Min<int> > BBIntMinNode;
//@}
/// @name Maximisation and integer evaluation
//@{
/// Node
typedef BBNode<int,Max<int> > BBIntMaxNode;
//@}
/// @name Minimization and long evaluation
//@{
/// Node
typedef BBNode<long,Min<long> > BBLongMinNode;
//@}
/// @name Maximisation and long evaluation
//@{
/// Node
typedef BBNode<long,Max<long> > BBLongMaxNode;
//@}
/// @name Minimization and double evaluation
//@{
/// Node
typedef BBNode<double,Min<double> > BBDoubleMinNode;
//@}
/// @name Maximisation and double evaluation
//@{
/// Node
typedef BBNode<double,Max<double> > BBDoubleMaxNode;
//@}
/// @name Minimization and float evaluation
//@{
/// Node
typedef BBNode<float,Min<float> > BBFloatMinNode;
//@}
/// @name Maximisation and float evaluation
//@{
/// Node
typedef BBNode<float,Max<float> > BBFloatMaxNode;
//@}

//@}
//@}
//
};
#endif



