
#ifndef BOB_ENVPROG
#define BOB_ENVPROG


namespace Bob {

/*------------------ EnvProg Class-------------*/
/**
 *  @defgroup EnvProgGRP The Bob++ environment
 *  @brief The Programming Environment
 *  @{
 *  This module groups the different programming environments proposed in
 *  Bob++.
 *
 * \page ProgEnvPage The Programming Environment
 *
 * \section AbsSect The Bob::AlgoEnvProg
 *  We try to explain here the way to add a new enviroment.
 *  The idea is to derive the abstract Bob::AlgoEnvProg and redefine
 *  the two operators void Bob::AlgoEnvProg::operator()(TheInstance *inst) and
 *  virtual void Bob::AlgoEnvProg::operator()(TheNode *nd,TheInstance *inst)
 *
 *  The first one is dedicated to the resolution of an instance given the
 *  parameter inst.
 *  The second one is dedicated to the resolution of a sub-space rooted on
 *  the node nd for the instance inst.
 *
 *  For a given environment, one could rewrite one
 *  Bob::AlgoEnvProg class for each of the algorithms
 *  Branch and Bound, Divide and Conquer....
 *
 * \section StaticSect Need of static environment class
 *  A specific environment could need a static class to call specific
 *  methods or function. For example in @ref AthaEnvProgGRP or
 *  @ref ThrEnvProgGRP, it has been necessary to add a specific class
 *  with static methods. Refer to this environment to know how to use them.
 *
 * \section DataSect The Data structures
 *
 * Now for a given parallelization, several Bob++ algorithms are spawned
 * on different processors to solve the same instance. The different algorithms
 * and the different processor need to share information or to communicate
 * information.
 * The way to do this in Bob++ is to derive the different data structures
 * used by the algorithms i.e. the Priority Queue Bob::PQ and the
 * Goal data structure which depends from each algorithm.
 *
 * For example, in the threaded environment (@ref ThrEnvProgGRP), to parallelize
 * a Branch and Bound, several Bob::BBAlgo are spawned in the different threads
 * These Bob::BBAlgo use \em wrapper on top of the Priority queue
 * and the solution data structure.
 * These \em wrappers have the same interfaces as the data structures,
 * but they use mutex to insure the mutual exculsion of
 * the operations.
 *
 * The Priority queue and the solution could be seen as high level
 * communication and/or synchronization tools.
 *
 * For the athapascan/kaapi environment (@ref AthaEnvProgGRP), the idea is
 * quite similar, in the sens that we also use the Priority queue to
 * generate parallelism. But in this case, the priority queue \em wrapper,
 * could insert a node on a "sequential" priority queue, or spawn a new
 * algorithm with this node as a new root.
 *
 */

/**
 *  @defgroup AbsEnvProgGRP The abstract class for the Programming environment
 *  @brief The base class for the programming environment for the Bob++ algorithms
 *  @{
 */

/** the Abstract AlgoEnvProg
 */
template<class Trait>
class AlgoEnvProg {
  typedef typename Trait::Instance TheInstance;
  typedef typename Trait::Node TheNode;
  typedef typename Trait::Goal TheGoal;
  typedef typename Trait::Stat TheStat;
  typedef typename Trait::PriComp ThePriComp;
  typedef PQSkew<TheNode, ThePriComp> ThePQ;
public:
  /// Constructor
  AlgoEnvProg() {
  }
  /// Destructor
  virtual ~AlgoEnvProg() {
  }
  /** solve the problem, given by the instance.
   * @param inst the Instance to solve
   */
  virtual void operator()(TheInstance *inst) = 0;
  /** solve the problem for the given instance with the root node nd
   * @param inst the Instance to solve
   * @param nd the root node to solve
   */
  virtual void operator()(TheNode *nd, TheInstance *inst) = 0;
  /** get the goal after a resolution
   * @return the goal with the solution.
   */
  virtual TheGoal *goal()=0;
  /** get the stat after a resolution
   * @return the stat.
   */
  virtual TheStat *stat()=0;
  /** print the stat according to the environment
   */
  virtual void display_stat()=0;
  /** print the stat according to the environment
   */
  virtual void display_goal()=0; 
  /** print the stat according to the environment
   */
  virtual void display() {
    display_stat();
    display_goal();
  }
};
/**
 * @}
 */

/**
 *  @defgroup SeqEnvProgGRP The sequential Programming environment
 *  @brief The base class for sequential programming environment for the Bob++ algorithms
 *  @{
 */


/** The Sequential AlgoEnvProg
 */
template<class Trait>
class SeqAlgoEnvProg : public AlgoEnvProg<Trait> {
  typedef typename Trait::Instance TheInstance;
  typedef typename Trait::Node TheNode;
  typedef typename Trait::Algo TheAlgo;
  typedef typename Trait::Goal TheGoal;
  typedef typename Trait::Stat TheStat;
  typedef typename Trait::PriComp ThePriComp;
  typedef PQSkew<TheNode, ThePriComp> ThePQ;
  typedef AlgoEnvProg<Trait> aep;

public:
  /** the constructor
   *  @param _d is an optional parameter to desable the display of stat, solution at the end of the resolution the defaut is display all !
   */
  SeqAlgoEnvProg(int _d = 3) : AlgoEnvProg<Trait>(), tg(0), st(0), pq(0), disp(_d) {
  }
  /// Destructor
  virtual ~SeqAlgoEnvProg() {
    if (tg != 0) delete tg;
    if (pq != 0) delete pq;
    if (st != 0) delete st;
  }
  /** solve the problem
    * @param inst the instance to solve
    */
  virtual void operator()(TheInstance *inst) {
    doit(0, inst);
  }
  /** solve the problem from the node \a nd
    * @param inst the instance to solve
    * @param nd the root node to solve
    */
  virtual void operator()(TheNode *nd, TheInstance *inst) {
    doit(nd, inst);
  }
  /** Returns the goal after a resolution
   * @return the goal with the solution.
   */
  TheGoal *goal() {
    return tg;
  }
  /** Returns the stat after a resolution
   * @return the stat.
   */
  TheStat *stat() {
    return st;
  }
  /** Displays the stat according to the environment
   */
  virtual void display_stat() {
    if ((disp&2) == 2) {
      st->PrtL(std::cout);
      st->PrtD(std::cout);
      pq->getStat()->PrtL(std::cout);
      pq->getStat()->PrtD(std::cout);
      tg->getStat()->PrtL(std::cout);
      tg->getStat()->PrtD(std::cout);
    }
  }
  /** Dispaly the Goal according to the environment
   */
  virtual void display_goal() {
    if ((disp&1) == 1) {
      tg->Prt(std::cout);
    } 
  }

protected:
  /** Do the job for the two interfaces
   * @param nd the node to search from.
   * @param inst the instance used to perform the search.
   */
  virtual void doit(TheNode *nd, TheInstance *inst) {
    pq = new PQSkew<TheNode, ThePriComp>();
    tg = new TheGoal();
    st = new TheStat(inst->id(), pq->getStat(), tg->getStat());
    st->namelog();
    inst->Init();
    tg->Init(inst);
    TheAlgo al(inst, pq, tg, st);
    if (nd == 0) inst->InitAlgo(&al);
    else al.GPQIns(nd);
    al();
    aep::display();
  }
protected:
  /// The Goal associated with the algorithm
  TheGoal *tg;
  /// The Stat associated with the algorithm
  TheStat *st;
  /// The priority Queue used in the algorithm
  PQSkew<TheNode, ThePriComp> *pq;
  /// Option to know the verbosity of the algorithm
  int disp;
};

/** The Sequential Branch and Bound AlgoEnvProg
 */
template<class Trait>
class SeqBBAlgoEnvProg : public SeqAlgoEnvProg<Trait> {
  typedef typename Trait::Instance TheInstance;
  typedef typename Trait::Node TheNode;
  typedef typename Trait::Algo TheAlgo;
  typedef typename Trait::Goal TheGoal;
  typedef typename Trait::Stat TheStat;
  typedef typename Trait::PriComp ThePriComp;

public:
  /// Constructor
  SeqBBAlgoEnvProg(int _d = 3) : SeqAlgoEnvProg<Trait>(_d) {
  }
  /// Destructor
  virtual ~SeqBBAlgoEnvProg() {}
};

/** The Sequential Simple Divide and Conquer AlgoEnvProg
 */
template<class Trait>
class SeqSDCAlgoEnvProg : public SeqAlgoEnvProg<Trait> {
  typedef typename Trait::Instance TheInstance;
  typedef typename Trait::Node TheNode;
  typedef typename Trait::Algo TheAlgo;
  typedef typename Trait::Goal TheGoal;
  typedef typename Trait::Stat TheStat;
  typedef typename Trait::PriComp ThePriComp;

public:
  /// Constructor
  SeqSDCAlgoEnvProg(int _d = 3) : SeqAlgoEnvProg<Trait>(_d) {
  }
  /// Destructor
  virtual ~SeqSDCAlgoEnvProg() {}
};


/** The Sequential Costed Divide and Conquer AlgoEnvProg
 */
template<class Trait>
class SeqCDCAlgoEnvProg : public SeqAlgoEnvProg<Trait> {
  typedef typename Trait::Instance TheInstance;
  typedef typename Trait::Node TheNode;
  typedef typename Trait::Algo TheAlgo;
  typedef typename Trait::Goal TheGoal;
  typedef typename Trait::Stat TheStat;
  typedef typename Trait::PriComp ThePriComp;

public:
  /// Constructor
  SeqCDCAlgoEnvProg(int _d=3) : SeqAlgoEnvProg<Trait>(_d) {
  }
  /// Denstructor
  virtual ~SeqCDCAlgoEnvProg() {}
};

/**
 * @}
 */

/**
 * @}
 */

}
#endif
