#ifndef BOBPRI
#define BOBPRI
/*---------------------------------------------------------------------*/

namespace Bob {
/*---------------------------------------------------------------------*/
/** @defgroup PriGRP The priority module to schedule all types of nodes
 *  @brief Base priority classes to schedule all types of nodes
 *  @ingroup SchedAlgoGRP
 *  When the goal of the algorithm is to perform a search in a space,
 *  the space could traversed by different way. The first step is then
 *  to allow to algorithm to schedule the search. We define here
 *  the first abstract classes for an associated prioity with node.
 */

/** the Simple priority just for test
 * @ingroup PriGRP
  */
template<class T>
struct SPri {
public:
  /**  is equal operator
   * @param n1 the first priority to compare
   * @param n2 the second priority to compare
   */
  static bool isEq(const T &n1, const T &n2) {
    return true;
  }
  /**  is better operator
   * @param n1 the first priority to compare
   * @param n2 the second priority to compare
   */
  static bool isBetter(const T &n1, const T &n2) {
    return true;
  }

};

/** the best first priority
 *
 * when the evaluations of two Nodes are equal, the selected Node is the
 * deepest one.
 * @ingroup PriGRP
  */
template<class T>
class BestDPri {
public:
  /**  is equal operator
  * @param e1 the first priority to compare
  * @param e2 the second priority to compare
  */
  static bool isEq(const T &e1, const T &e2) {
    return e1 == e2 && e1.dist() == e2.dist();
  }
  /**  is better operator
   * @param e1 the first priority to compare
   * @param e2 the second priority to compare
   */
  static bool isBetter(const T &e1, const T &e2) {
    return (e1 > e2 ? 1 : (e1 == e2 ? e1.dist() > e2.dist() : 0));
  }

};

/** the best first priority
 *
 * when the evaluations of two Nodes are equal, the selected Node is the
 * shallowest one.
 * @ingroup PriGRP
  */
template<class T>
class BestEPri {
public:
  /**  is equal operator
  * @param e1 the first priority to compare
  * @param e2 the second priority to compare
  */
  static bool isEq(const T &e1, const T &e2) {
    return e1 == e2 && e1.dist() == e2.dist() ;
  }
  /**  is equal operator
  * @param e1 the first priority to compare
  * @param e2 the second priority to compare
  */
  static bool isBetter(const T &e1, const T &e2) {
    return (e1 > e2 ? 1 : (e1 == e2 ? e1.dist() < e2.dist() : 0));
  }

};

/** the Depth first priority
 * @ingroup PriGRP
  */
template<class T>
class DepthPri {
public:
  /**  is equal operator
  * @param e1 the first priority to compare
  * @param e2 the second priority to compare
  */
  static bool isEq(const T &e1, const T &e2) {
    return e1.dist() == e2.dist();
  }
  /**  is equal operator
  * @param e1 the first priority to compare
  * @param e2 the second priority to compare
  */
  static bool isBetter(const T &e1, const T &e2) {
    return e1.dist() > e2.dist();
  }

};

/** A second Depth first priority
 * Depth First priority but when the depth of the nodes are equal, we get
 * the best one.
 * @ingroup PriGRP
  */
template<class T>
class DepthEPri {
public:
  /**  is equal operator
  * @param e1 the first priority to compare
  * @param e2 the second priority to compare
  */
  static bool isEq(const T &e1, const T &e2) {
    return e1.dist() == e2.dist() && e1 == e2;
  }
  /**  is equal operator
  * @param e1 the first priority to compare
  * @param e2 the second priority to compare
  */
  static bool isBetter(const T &e1, const T &e2) {
    return e1.dist() > e2.dist() || (e1.dist()==e2.dist() && e1>e2);
  }

};

/** the best first, then depth first priority
 * @ingroup PriGRP
  */
template<class T>
class BestThenDepthPri {
public:
  /**  is equal operator
  * @param e1 the first priority to compare
  * @param e2 the second priority to compare
  */
  static bool isEq(const T &e1, const T &e2) {
    if( e1.dist() >= core::BTDthreshold() && e2.dist() >= core::BTDthreshold() )
      return e1.dist() == e2.dist() && e1 == e2;
    else
      return e1 == e2;
  }
  /**  is equal operator
  * @param e1 the first priority to compare
  * @param e2 the second priority to compare
  */
  static bool isBetter(const T &e1, const T &e2) {
    if( e1.dist() >= core::BTDthreshold() && e2.dist() >= core::BTDthreshold() )
      return e1.dist() > e2.dist()|| (e1.dist()==e2.dist() && e1>e2);
    else
      return e1 > e2;
  }

};
};

#endif
